from math import ceil, log2
from secrets import token_bytes
from typing import Callable, Optional, Union

from puid import Chars
from puid.bits import muncher
from puid.chars import CustomChars, PredefinedChars, ValidChars
from puid.chars_error import InvalidChars
from puid.encoder import encoder
from puid.entropy import bits_for_total_risk
from puid.puid_error import BitsError, TotalRiskError


class Puid:
    __slots__ = (
        'chars',
        'len',
        'bits',
        'entropy_source',
        'bits_muncher',
        'bits_per_char',
        '_encoded',
        'ere',
    )

    def __init__(
        self,
        total: Optional[Union[int, float]] = None,
        risk: Optional[Union[int, float]] = None,
        bits: Optional[int] = None,
        chars: Optional[Union[Chars, str]] = None,
        entropy_source: Optional[Callable[[int], bytearray]] = None,
    ) -> None:

        base_bits = None
        if bits is None and total is None and risk is None:
            base_bits = 128
        elif bits is not None and (total is not None or risk is not None):
            raise BitsError("bits cannot be specified with total/risk")
        elif bits is not None and total is None and risk is None:
            if not isinstance(bits, int):
                raise BitsError("bits must be non-negative integer")
            if bits <= 0:
                raise BitsError("bits must be non-negative integer")
            base_bits = bits
        elif total is None and risk is not None:
            raise TotalRiskError("risk with no total")
        elif total is not None and risk is None:
            raise TotalRiskError("total with no risk")
        else:
            base_bits = bits_for_total_risk(total, risk)

        if chars is None:
            self.chars: ValidChars = PredefinedChars(Chars.SAFE64)
        elif isinstance(chars, Chars):
            self.chars = PredefinedChars(chars)
        elif isinstance(chars, str):
            self.chars = CustomChars(chars)
        else:
            raise InvalidChars('specified chars must be either Chars enum or string')

        n_chars = len(self.chars)
        n_bits_per_char = log2(n_chars)
        self.len: int = round(ceil(base_bits / n_bits_per_char))
        self.bits: float = self.len * n_bits_per_char
        entropy_fn = entropy_source or token_bytes
        self.entropy_source: str = f'{entropy_fn.__module__}.{entropy_fn.__name__}'

        self.bits_muncher: Callable[[], list[int]] = muncher(n_chars, self.len, entropy_fn)

        self.bits_per_char: float = n_bits_per_char

        chars_encoder = encoder(self.chars)

        def encoded(values: list[int]) -> list[str]:
            return [chr(chars_encoder(value)) for value in values]

        self._encoded: Callable[[list[int]], list[str]] = encoded

        self.ere: float = (n_bits_per_char * n_chars) / (8 * len(self.chars.value.encode('utf-8')))

    def __repr__(self) -> str:
        bits = round(self.bits, 2)
        bpc = round(self.bits_per_char, 2)
        return f'Puid: bits = {bits}, bits_per_char = {bpc}, chars = {self.chars}, len = {self.len}, '
        'ere = {self.ere}, entropy_source = {self.entropy_source}'

    def generate(self) -> str:
        values = self.bits_muncher()
        return "".join(self._encoded(values))



def safe32():
    two = ord("2")
    six = ord("6") - 3
    b = ord("b")
    d = ord("d")
    f = ord("f") - 9
    j = ord("j")
    m = ord("m") - 13
    p = ord("p") - 15
    t = ord("t")
    B = ord("B")
    D = ord("D")
    F = ord("F") - 21
    J = ord("J")
    L = ord("L") - 25
    P = ord("P") - 28
    T = ord("T")

    def encoder(n):
        if n < 3:
            return n + two
        if n < 7:
            return n + six
        if n == 7:
            return b
        if n == 8:
            return d
        if n < 12:
            return n + f
        if n == 12:
            return j
        if n < 15:
            return n + m
        if n < 18:
            return n + p
        if n == 18:
            return t
        if n == 19:
            return B
        if n == 20:
            return D
        if n < 24:
            return n + F
        if n == 24:
            return J
        if n < 28:
            return n + L
        if n < 31:
            return n + P
        return T

    return encoder



class PuidError(Exception):
    """Base class for Puid Errors"""

    pass


class BitsError(PuidError):
    """
    Raised when
      - bits is non-negative
      - bits specified with total/risk
    """
    pass


class TotalRiskError(PuidError):
    """
    Raised when
      - total or risk are not non-negative
      - total and risk are not both specified
    """
    pass



def _hex_case(upper):
    decimal = ord("0")
    alpha = (ord("A") if upper else ord("a")) - 10

    def encoder(n):
        if n < 10:
            return n + decimal
        return n + alpha

    return encoder


def hex_lower():
    return _hex_case(False)


def hex_upper():
    return _hex_case(True)



class CharsError(Exception):
    """Base class for Chars Errors"""

    pass


class InvalidChars(CharsError):
    """Raised when characters string contains invalid character"""

    pass


# class NotValidChars(CharsError):
#     """Raised when value is not an instance of ValidChars"""

#     pass


class NonUniqueChars(CharsError):
    """Raised when characters string contains duplicate characters"""

    pass


class TooFewChars(CharsError):
    """Raised when characters string contains too few characters"""

    pass


class TooManyChars(CharsError):
    """Raised when characters string contains too many characters"""

    pass



def safe_ascii():
    bang = ord("!")
    ampersand = ord("&") - 4
    openSquareBracket = ord("[") - 56
    underscore = ord("_") - 59
    a = ord("a") - 60
    tilde = ord("~")

    def encode(n):
        if n == 0:
            return bang
        if n < 5:
            return n + ampersand
        if n < 57:
            return n + openSquareBracket
        if n < 60:
            return n + underscore
        if n < 89:
            return n + a
        return tilde

    return encode



def safe64():
    upper = ord("A")
    lower = ord("a") - 26
    decimal = ord("0") - 52
    hyphen = ord("-")
    underscore = ord("_")

    def encoder(n):
        if n < 26:
            return n + upper
        if n < 52:
            return n + lower
        if n < 62:
            return n + decimal
        if n == 62:
            return hyphen
        return underscore

    return encoder



def alpha():
    upper = ord("A")
    lower = ord("a") - 26

    return lambda n: n + (upper if n < 26 else lower)


def alpha_lower():
    return _alpha_case(False)


def alpha_upper():
    return _alpha_case(True)


def _alpha_case(upper):
    alpha = ord("A") if upper else ord("a")
    return lambda n: n + alpha



from typing import Callable, Dict

from puid.chars import Chars, ValidChars
from puid.encoders.alpha import alpha
from puid.encoders.alpha import alpha_lower
from puid.encoders.alpha import alpha_upper
from puid.encoders.alphanum import alphanum
from puid.encoders.alphanum import alphanum_lower
from puid.encoders.alphanum import alphanum_upper
from puid.encoders.base16 import base16
from puid.encoders.base32 import base32
from puid.encoders.base32 import base32_hex
from puid.encoders.base32 import base32_hex_upper
from puid.encoders.crockford32 import crockford32
from puid.encoders.custom import custom
from puid.encoders.decimal import decimal
from puid.encoders.hex import hex_lower
from puid.encoders.hex import hex_upper
from puid.encoders.safe32 import safe32
from puid.encoders.safe64 import safe64
from puid.encoders.safe_ascii import safe_ascii
from puid.encoders.symbol import symbol
from puid.encoders.word_safe32 import word_safe32


_ENCODER_CACHE: Dict[str, Callable[[int], int]] = {}


def _init_encoder_map() -> Dict[str, Callable[[int], int]]:
    return {
        Chars.ALPHA.name: alpha(),
        Chars.ALPHA_LOWER.name: alpha_lower(),
        Chars.ALPHA_UPPER.name: alpha_upper(),
        Chars.ALPHANUM.name: alphanum(),
        Chars.ALPHANUM_LOWER.name: alphanum_lower(),
        Chars.ALPHANUM_UPPER.name: alphanum_upper(),
        Chars.BASE16.name: base16(),
        Chars.BASE32.name: base32(),
        Chars.BASE32_HEX.name: base32_hex(),
        Chars.BASE32_HEX_UPPER.name: base32_hex_upper(),
        Chars.CROCKFORD32.name: crockford32(),
        Chars.DECIMAL.name: decimal(),
        Chars.HEX.name: hex_lower(),
        Chars.HEX_UPPER.name: hex_upper(),
        Chars.SAFE32.name: safe32(),
        Chars.SAFE64.name: safe64(),
        Chars.SAFE_ASCII.name: safe_ascii(),
        Chars.SYMBOL.name: symbol(),
        Chars.WORD_SAFE32.name: word_safe32(),
    }


def encoder(chars: ValidChars) -> Callable[[int], int]:
    if not _ENCODER_CACHE:
        _ENCODER_CACHE.update(_init_encoder_map())
    
    encoder_fn = _ENCODER_CACHE.get(chars.name)
    if encoder_fn is not None:
        return encoder_fn
    return custom(chars)



from collections import namedtuple
from math import ceil, floor, log2

AcceptResult = namedtuple('AcceptResult', ['accept', 'shift'])

#  Create array of minimum bits required to determine if a value is less than n_chars
#  Array elements are of the form (n, bits): For values less than n, bits bits are required
#
#  As example, the bits shifts array for the 36 AlphaNumLower characters is:
#    (36, 6), (39, 5), (47, 3), (63, 2)
#
#  Each value slice uses 6 bits of entropy. In bits, 36 is 100100.
#  Now suppose we slice the value 50. In bits, 50 is 110010.
#
#  Only two bits are necessary to determine 100100 < 110010
#


def bit_shifts(n_chars):
    n_bits_per_char = ceil(log2(n_chars))

    base_value = n_chars - 1 if n_chars % 2 == 0 else n_chars
    base_shift = (base_value, n_bits_per_char)

    if log2(n_chars).is_integer():
        return [base_shift]

    def is_bit_zero(bit):
        return base_value & (1 << (bit - 1)) == 0

    def pow2(bit):
        return round(pow(2, bit))

    def shift(bit):
        return (base_value | pow2(bit) - 1, n_bits_per_char - bit + 1)

    return [base_shift] + [shift(bit) for bit in range(2, n_bits_per_char) if is_bit_zero(bit)]


def fill_entropy(entropy_offset, entropy_bytes, entropy_fn):
    n_bytes = len(entropy_bytes)
    n_bits = 8 * n_bytes

    if entropy_offset == n_bits:
        # no carry
        entropy_bytes[0:n_bytes] = entropy_fn(n_bytes)
    else:
        offset_byte_num = floor(entropy_offset / 8)

        # Move unused bytes to the left
        entropy_bytes[0 : n_bytes - offset_byte_num] = entropy_bytes[offset_byte_num:n_bytes]

        # Fill right bytes with new random values
        entropy_bytes[n_bytes - offset_byte_num : n_bytes] = entropy_fn(offset_byte_num)

    return entropy_offset % 8


def value_at(l_offset, n_bits, bytes):
    l_byte_ndx = floor(l_offset / 8)
    l_byte = bytes[l_byte_ndx]

    l_bit_num = l_offset % 8

    if l_bit_num + n_bits <= 8:
        return ((l_byte << l_bit_num) & 0xFF) >> (8 - n_bits)

    r_byte = bytes[l_byte_ndx + 1]
    r_bit_num = l_bit_num + n_bits - 8

    l_value = ((l_byte << l_bit_num) & 0xFF) >> (l_bit_num - r_bit_num)
    r_value = r_byte >> (8 - r_bit_num)

    return l_value + r_value


def muncher(n_chars, puid_len, entropy_fn):
    n_bits_per_char = ceil(log2(n_chars))
    n_bits_per_puid = n_bits_per_char * puid_len
    n_bytes_per_puid = ceil(n_bits_per_puid / 8)

    buffer_len = n_bytes_per_puid + 1
    n_entropy_bits = 8 * buffer_len
    entropy_offset = n_entropy_bits
    entropy_bytes = bytearray(buffer_len)

    def pow2(bit):
        return round(pow(2, bit))

    def is_pow2(n):
        return pow2(round(log2(n))) == n

    counter = list(range(puid_len))

    def sliced_value():
        nonlocal entropy_offset
        if n_entropy_bits < entropy_offset + n_bits_per_char:
            entropy_offset = fill_entropy(entropy_offset, entropy_bytes, entropy_fn)
        return value_at(entropy_offset, n_bits_per_char, entropy_bytes)

    if is_pow2(n_chars):
        #  When chars count is a power of 2, sliced bits always yield a valid value
        def bits_muncher():
            def slice_value():
                nonlocal entropy_offset
                value = sliced_value()
                entropy_offset += n_bits_per_char
                return value

            puid = [slice_value() for _ in counter]
            return puid

        return bits_muncher

    shifts = bit_shifts(n_chars)

    def accept_value(value):
        # Value is valid if it is less than the number of characters
        if value < n_chars:
            return AcceptResult(True, n_bits_per_char)

        # For invalid value, shift the minimal bits necessary to determine validity
        # If only one item, no need to search
        if len(shifts) == 1:
            return AcceptResult(False, shifts[0][1])

        bit_shift = [bs for bs in shifts if value <= bs[0]]
        return AcceptResult(False, bit_shift[0][1])

    def slice_value():
        nonlocal entropy_offset
        value = sliced_value()
        accept, shift = accept_value(value)
        # Returned shift is the minimal bits necessary to determine if slice value is valid
        entropy_offset += shift

        if accept:
            return value

        # If value not acceptable, slice another
        return slice_value()

    def bits_muncher():
        return [slice_value() for _ in counter]

    return bits_muncher



# n: 01234567 8 901 2 3 456 789 0 123 4 5 678 901
# c: 23456789 C FGH J M PQR VWX c fgh j m pqr vwx

def word_safe32():
    def encoder(n):
        two = ord("2")
        C = ord("C")
        F = ord("F")
        J = ord("J")
        M = ord("M")
        P = ord("P")
        V = ord("V")
        c = ord("c")
        f = ord("f")
        j = ord("j")
        m = ord("m")
        p = ord("p")
        v = ord("v")

        if (n < 8):
            return n + two
        if (n == 8):
            return C
        if (n < 12):
            return n - 9 + F
        if (n == 12):
            return J
        if (n == 13):
            return M
        if (n < 17):
            return n - 14 + P
        if (n < 20):
            return n - 17 + V
        if (n == 20):
            return c
        if (n < 24):
            return n - 21 + f
        if (n == 24):
            return j
        if (n == 25):
            return m
        if (n < 29):
            return n - 26 + p
        return n - 29 + v

    return encoder



def base32():
    alpha = ord("A")
    decimal = ord("2") - 26

    def encoder(n):
        return n + (alpha if n < 26 else decimal)

    return encoder


def base32_hex():
    return _base32_hex_case(False)


def base32_hex_upper():
    return _base32_hex_case(True)


def _base32_hex_case(upper):
    decimal = ord("0")
    alpha = (ord("A") if upper else ord("a")) - 10

    def encoder(n):
        return n + (decimal if n < 10 else alpha)

    return encoder
